@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  *
  *  BlueZ - Bluetooth protocol stack for Linux
@@ -6,20 +7,6 @@
  *  Copyright (C) 2002-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU Lesser General Public
- *  License as published by the Free Software Foundation; either
- *  version 2.1 of the License, or (at your option) any later version.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -52,7 +39,7 @@ static void print_handle(uint16_t handle)
 
 static void print_rssi(int8_t rssi)
 {
-	packet_print_rssi(rssi);
+	packet_print_rssi("RSSI", rssi);
 }
 
 static void print_sco_routing(uint8_t routing)
@@ -707,7 +694,13 @@ void broadcom_lm_diag(const void *data, uint8_t size)
 	}
 }
 
+static void lm_diag_evt(const void *data, uint8_t size)
+{
+	broadcom_lm_diag(data, 63);
+}
+
 static const struct vendor_evt vendor_evt_table[] = {
+	{ 0xb4, "LM Diag", lm_diag_evt, 64, true },
 	{ }
 };
 
